from enum import Enum
from typing import Union, List, Optional, Any
from myko import MItem, MCommand, Schema
from .utils import generate_hash

class Target(MItem):
    def __init__(self, id: str, name: str,
                 sub_targets: List[str], parent_targets: List[str], service_id: str,
                 bg_color: str, fg_color: str, last_updated: str, category: str, root_level: bool):
        super().__init__(id, name)

        self.category = category
        self.bgColor = bg_color
        self.fgColor = fg_color
        self.parentTargets = parent_targets
        self.subTargets = sub_targets
        self.serviceId = service_id
        self.lastUpdated = last_updated
        self.rootLevel = root_level
        # self.hash = str(uuid())


class TargetStatusEnum(Enum):
    Online = 'online'
    Offline = 'offline'


class TargetStatus(MItem):
    def __init__(self, id: str, name: str, target_id: str, status: TargetStatusEnum, last_updated: str, instance_id: str):
        super().__init__(id, name)

        self.targetId = target_id
        self.status = status.value
        self.lastUpdated = last_updated
        self.instanceId = instance_id
        # self.hash = generate_hash()


class Action(MItem):
    def __init__(self, id: str, name: str, schema: Optional[Schema], target_id: str, service_id: str):
        super().__init__(id, name)

        self.schema = schema
        self.targetId = target_id
        self.serviceId = service_id
        # self.hash = generate_hash()


class Emitter(MItem):
    def __init__(self, id: str, name: str, schema: Optional[Schema], target_id: str, service_id: str):
        super().__init__(id, name)

        self.schema = schema
        self.targetId = target_id
        self.serviceId = service_id
        # self.hash = generate_hash()


class Pulse(MItem):
    def __init__(self, name: str, emitter_id: str, data: Any):
        super().__init__(emitter_id, name)

        self.emitterId = emitter_id
        self.data = data
        # self.hash = generate_hash()


class KeyframeData():
    type: str

class NumberKeyframeData(KeyframeData):
    def __init__(self, value: float, fade_in: Optional[float] = None, curve_id: Optional[str] = None):
        self.type = 'number'
        self.value = value
        self.fade_in = fade_in
        self.curve_id = curve_id

class IntegerKeyframeData(KeyframeData):
    def __init__(self, value: int, fade_in: Optional[float] = None, curve_id: Optional[str] = None):
        self.type = 'integer'
        self.value = value
        self.fade_in = fade_in
        self.curve_id = curve_id



class StringKeyframeData(KeyframeData):
    def __init__(self, value: str):
        self.type = 'string'
        self.value = value

class BooleanKeyframeData(KeyframeData):
    def __init__(self, value: bool):
        self.type = 'boolean'
        self.value = value


class GlobalVariableKeyframeData(KeyframeData):
    def __init__(self, value: str, fade_in: Optional[float] = None, curve_id: Optional[str] = None):
        self.type = 'global_variable'
        self.value = value
        self.fade_in = fade_in
        self.curve_id = curve_id


class SceneOutputKeyframeData(KeyframeData):
    def __init__(self, value: str, fade_in: Optional[float] = None, curve_id: Optional[str] = None):
        self.type = 'scene_output'
        self.value = value
        self.fade_in = fade_in
        self.curve_id = curve_id


class NullKeyframeData(KeyframeData):
    def __init__(self):
        self.type = 'null'
        self.value = None

    


class TimeMode():
    type: str


class ClockTimeMode(TimeMode):
    def __init__(self):
        self.type = 'clock'


class BeatTimeMode(TimeMode):
    def __init__(self, measure: int):
        self.type = 'beat'
        self.measure = measure


class FrameTimeMode(TimeMode):
    def __init__(self, rate: float):
        self.type = 'frame'
        self.rate = rate


class EventTrack(MItem):
    def __init__(self, id: str, name: str, time_mode: TimeMode, service_id: str):
        super().__init__(id, name)
        self.timeMode = time_mode
        self.sourceMode = {'type': 'sourced', 'service_id': service_id}


class ConcreteSchemaType(Enum):
    Null = 'null'
    String = 'string'
    Number = 'number'
    Integer = 'integer'
    Boolean = 'boolean'
    Object = 'object'
    Array = 'array'
    Ref = 'ref'
    

class EventTrackLane(MItem):
    def __init__(self, id: str, name: str, event_track_id: str, type: ConcreteSchemaType):
        super().__init__(id, name)
        self.eventTrackId = event_track_id
        self.type = type.value
        

class Keyframe(MItem):
    def __init__(self, id: str, event_track_id: str, layer_id: str, time: int, data: KeyframeData):
        super().__init__(id, f"{time}")
        self.eventTrackId = event_track_id
        self.layerId = layer_id
        self.time = time
        self.data = data


class InstanceStatusEnum(Enum):
    Starting = 'Starting'
    Available = 'Available'
    Stopping = 'Stopping'
    Unavailable = 'Unavailable'
    Error = 'Error'


class Instance(MItem):
    def __init__(self, id: str, name: str, service_id: str, cluster_id: str, service_type_code: str,
                 status: InstanceStatusEnum, machine_id: str, color: str, message: str):
        super().__init__(id, name)

        self.serviceId = service_id
        self.serviceTypeCode = service_type_code
        self.clusterId = cluster_id
        self.machineId = machine_id
        self.color = color
        # self.hash = generate_hash()
        
        self.message = message
        self.status = status.value


class Machine(MItem):
    def __init__(self, name: str):
        super().__init__(name, name)

class ExecTargetAction(MCommand):
    def __init__(self, tx: str, createdAt: str, action: Action, data: any):
        super().__init__(tx, createdAt)

        self.action = action
        self.data = data
        self.hash = generate_hash()


class AlertLevel(Enum):
    INFO = 'info'
    WARN = 'warn'
    ERROR = 'error'


class AlertEntityType(Enum):
    TARGET = 'Target'
    ACTION = 'Action'
    PAYLOAD = 'Payload'
    INSTANCE = 'Instance'


class Alert(MItem):
    def __init__(self, entity_id: str, entity_type: AlertEntityType, instance_id: str,
                 level: AlertLevel, message: str, code: str):
        super().__init__(entity_id + instance_id, f"{entity_id}:{code}")
        self.entityId = entity_id
        self.entityType = entity_type.value
        self.instanceId = instance_id
        self.level = level.value
        self.message = message
        self.code = code
        # self.hash = generate_hash()


from .client import RshipExecClient
from .proxies.instance import InstanceProxy, InstanceArgs
from .proxies.target import TargetProxy, TargetArgs
from .proxies.emitter import EmitterProxy, EmitterArgs
from .proxies.action import ActionProxy, ActionArgs
from .proxies.event_track import EventTrackProxy, EventTrackArgs, EventTrackLaneProxy, EventTrackLaneArgs, KeyframeArgs
from .models import TimeMode, BeatTimeMode, ClockTimeMode, FrameTimeMode, ConcreteSchemaType, KeyframeData, NumberKeyframeData, IntegerKeyframeData, StringKeyframeData, GlobalVariableKeyframeData, SceneOutputKeyframeData, NullKeyframeData


from ..models import EventTrack, EventTrackLane, TimeMode, Keyframe, ConcreteSchemaType, KeyframeData
from typing import List, Dict
from uuid import uuid4

class EventTrackArgs:
    def __init__(self, name: str, short_id: str, time_mode: TimeMode):
        self.name = name
        self.short_id = short_id
        self.time_mode = time_mode


class KeyframeArgs:
    def __init__(self, time: float, data: KeyframeData):
        self.time = time
        self.data = data
        

class EventTrackLaneArgs:
    def __init__(self, short_id: str, name: str, type: ConcreteSchemaType, keys: List[KeyframeArgs]):
        self.short_id = short_id
        self.name = name
        self.type = type
        self.keys = keys


class EventTrackProxy:
    def __init__(self, instance, args: EventTrackArgs, client):
        self.client = client
        self.args = args
        self.instance = instance
        self.lanes: Dict[str, EventTrackLaneProxy] = {}
   
    def id(self) -> str:
        return f"{self.instance.args.service_id}:{self.args.short_id}"
    
    async def add_lane(self, args: EventTrackLaneArgs) -> 'EventTrackLaneProxy':
        p = EventTrackLaneProxy(self.instance, self.client, args, self)
        self.lanes[p.id()] = p
        await self.save() # save EventTrackProxy
        await p.save() # save EventTrackLaneProxy
        return p
    
    async def save(self):
        event_track = EventTrack(
            id=self.id(),
            name=self.args.name,
            time_mode=self.args.time_mode.__dict__,
            service_id=self.instance.args.service_id,
        )

        if event_track.id not in self.client.event_tracks or self.client.event_tracks[event_track.id] != event_track.hash:
            await self.client.set_data(event_track, 'EventTrack')
            self.client.event_tracks[event_track.id] = event_track

    async def save_all(self):
        await self.save()
        for lane in self.lanes.values():
            await lane.save()


class EventTrackLaneProxy:
    def __init__(self, instance, client, args: EventTrackLaneArgs, track: EventTrackProxy):
        self.client = client
        self.args = args
        self.track = track
        self.sent = {}

    def id(self) -> str:
        return f"{self.track.id()}:{self.args.short_id}"
    
    async def save(self):
        tx = str(uuid4())

        frames = [
            Keyframe(
                id=f"{self.id()}:{key.time}",
                event_track_id=self.track.id(),
                layer_id=self.id(),
                time=key.time,
                data=key.data.__dict__,
            )
            for key in self.args.keys
        ]

        grouped_by_lanes = {}
        for frame in frames:
            if frame.layerId not in grouped_by_lanes:
                grouped_by_lanes[frame.layerId] = []
            grouped_by_lanes[frame.layerId].append(frame)

        lanes = [
            EventTrackLane(
                id=layer_id,
                name=self.args.name,
                event_track_id=self.track.id(),
                type=self.args.type,
            )
            for layer_id in grouped_by_lanes
        ]

        for lane in lanes:
            await self.client.set_data(lane, 'EventTrackLayer')

        to_delete = dict(self.sent)
        for frame in frames:
            if frame.id in to_delete:
                to_delete.pop(frame.id)

        self.sent.clear()

        for frame in frames:
            self.sent[frame.id] = frame
            await self.client.set_data(frame, 'Keyframe')
        
        for frame_id, frame in to_delete:
            await self.client.delete_data(frame, 'Keyframe')

        self.sent.clear()


import json
import websockets
import traceback

from typing import Any, Callable, Dict, Coroutine
from myko import MEvent, MEventType, MCommand, WSMEvent, WSMCommand

from .models import Target, Action, Emitter, Pulse,  Machine, Alert, InstanceStatusEnum, TargetStatus, EventTrack
from .utils import generate_hash, get_current_timestamp

from .proxies.instance import InstanceProxy, InstanceArgs

class RshipExecClient:
    def __init__(self, rship_host: str, rship_port: int):
        self.rship_host = rship_host
        self.rship_port = rship_port
        self.is_connected = False
        self.websocket = None
        self.targets: Dict[str, Target] = {}
        self.target_statuses: Dict[str, TargetStatus] = {}
        self.actions: Dict[str, Action] = {}
        self.emitters: Dict[str, Emitter] = {}
        self.instances: Dict[str, InstanceProxy] = {}
        self.machines: Dict[str, Machine] = {}
        self.alerts: Dict[str, Alert] = {}
        self.handlers: Dict[str, Coroutine[Action, Any, None]] = {}
        self.event_tracks: Dict[str, EventTrack] = {}

    ##############################
    # Websockets
    ##############################

    async def connect(self):
        uri = f"ws://{self.rship_host}:{self.rship_port}/myko"
        print(f"Attempting to connect to {uri}")
        try:
            self.websocket = await websockets.connect(uri)
            print("Connected to Rship server successfully.")
            self.is_connected = True
        except Exception as e:
            print(f"Failed to connect to rship - {e}")

    async def disconnect(self):
        if self.websocket:
            await self.websocket.close()
            self.websocket = None
            print("Disconnected from Rship server")
            self.is_connected = False

    async def send_event(self, event: MEvent):
        try: 
            await self.websocket.send(json.dumps(WSMEvent(event).__dict__))
        except TypeError as e:
            print(f"Failed to send event: TypeError - {e}")
            traceback.print_exc()

    async def send_command(self, command: MCommand):
        try:
            await self.websocket.send(json.dumps(WSMCommand(command).__dict__))
        except ConnectionError as e:
            print(f"Failed to send command: ConnectionError - {e}")

    async def receive_messages(self):
        while self.websocket:
            try:
                message = await self.websocket.recv()
                # print(f"Received message: {message}")
                await self.handle_message(message)
            except websockets.ConnectionClosed:
                await self.disconnect()

    async def handle_message(self, message: str):
        data = json.loads(message)
        # print(f"Received command: {data}")
        if data["event"] == "ws:m:command":
            command_data = data["data"]
            command_id = command_data["commandId"]
            if command_id == "ExecTargetAction":
                action_id = command_data["command"]["action"]["id"]
                if action_id in self.actions:
                    action = self.actions[action_id]
                    handler = self.handlers.get(action_id)
                    if handler:
                       await handler(action, command_data["command"]["data"])

    ##############################
    # RSHIP API
    ##############################

    async def set_data(self, item: Any, item_type: str):
        event = MEvent(item=item, item_type=item_type, change_type=MEventType.SET,
                       created_at=get_current_timestamp(), tx=generate_hash())
        await self.send_event(event)


    async def delete_data(self, item: Any, item_type: str):
        event = MEvent(item=item, item_type=item_type, change_type=MEventType.DEL,
                       created_at=get_current_timestamp(), tx=generate_hash())
        await self.send_event(event)


    async def add_instance(self, args: InstanceArgs):
        instance_proxy = InstanceProxy(args=args, 
                                       client=self)
        await instance_proxy.save(InstanceStatusEnum.Available)

        self.instances[instance_proxy.id()] = instance_proxy
        return instance_proxy


    async def pulse_emitter(self, service_short_id: str, target_short_id: str, emitter_short_id: str, data: Any):
        full_emitter_id = f"{service_short_id}:{target_short_id}:{emitter_short_id}"
        pulse = Pulse(name="", emitter_id=full_emitter_id, data=data)
        await self.set_data(pulse, 'Pulse')
